@@ -660,6 +660,9 @@ void GraphicsContext::fillPath(const Path& path)
     if (paintingDisabled())
         return;
 
+    if (!(m_state.fillPattern || m_state.fillGradient || m_state.fillColor.isValid()))
+        return;
+
     QPainter* p = m_data->p();
     QPainterPath platformPath = path.platformPath();
     platformPath.setFillRule(toQtFillRule(fillRule()));
@@ -824,6 +827,9 @@ void GraphicsContext::fillRect(const FloatRect& rect)
     if (paintingDisabled())
         return;
 
+    if (!(m_state.fillPattern || m_state.fillGradient || m_state.fillColor.isValid()))
+        return;
+
     QPainter* p = m_data->p();
     QRectF normalizedRect = rect.normalized();
 
@@ -1725,6 +1731,10 @@ void GraphicsContext::takeOwnershipOfPlatformContext()
     m_data->takeOwnershipOfPlatformContext();
 }
 
+bool GraphicsContext::isAcceleratedContext() const
+{
+    return (platformContext()->paintEngine()->type() == QPaintEngine::OpenGL2);
 }
 
+}
 // vim: ts=4 sw=4 et
