@@ -211,10 +211,13 @@ QWebPageAdapter::QWebPageAdapter()
     , forwardUnsupportedContent(false)
     , insideOpenCall(false)
     , clickCausedFocus(false)
+    , mousePressed(false)
     , m_useNativeVirtualKeyAsDOMKey(false)
     , m_totalBytes(0)
     , m_bytesReceived()
     , networkManager(0)
+    , m_deviceOrientationClient(0)
+    , m_deviceMotionClient(0)
 {
     WebCore::initializeWebCoreQt();
 }
@@ -247,18 +250,19 @@ void QWebPageAdapter::initializeWebCorePage()
 
 #if ENABLE(DEVICE_ORIENTATION)
     if (useMock) {
-        DeviceOrientationClientMock* mockOrientationClient = new DeviceOrientationClientMock;
-        WebCore::provideDeviceOrientationTo(page, mockOrientationClient);
-
-        DeviceMotionClientMock* mockMotionClient= new DeviceMotionClientMock;
-        WebCore::provideDeviceMotionTo(page, mockMotionClient);
+        m_deviceOrientationClient = new DeviceOrientationClientMock;
+        m_deviceMotionClient = new DeviceMotionClientMock;
     }
 #if HAVE(QTSENSORS)
     else {
-        WebCore::provideDeviceOrientationTo(page, new DeviceOrientationClientQt);
-        WebCore::provideDeviceMotionTo(page, new DeviceMotionClientQt);
+        m_deviceOrientationClient =  new DeviceOrientationClientQt;
+        m_deviceMotionClient = new DeviceMotionClientQt;
     }
 #endif
+    if (m_deviceOrientationClient)
+        WebCore::provideDeviceOrientationTo(page, m_deviceOrientationClient);
+    if (m_deviceMotionClient)
+        WebCore::provideDeviceMotionTo(page, m_deviceMotionClient);
 #endif
 
     // By default each page is put into their own unique page group, which affects popup windows
@@ -270,7 +274,7 @@ void QWebPageAdapter::initializeWebCorePage()
 
     page->addLayoutMilestones(DidFirstVisuallyNonEmptyLayout);
 
-    settings = new QWebSettings(page->settings());
+    settings = new QWebSettings(page->settings(), page->group().groupSettings());
 
 #if ENABLE(NOTIFICATIONS) || ENABLE(LEGACY_NOTIFICATIONS)
     WebCore::provideNotification(page, NotificationPresenterClientQt::notificationPresenter());
@@ -289,6 +293,10 @@ QWebPageAdapter::~QWebPageAdapter()
 #if ENABLE(NOTIFICATIONS) || ENABLE(LEGACY_NOTIFICATIONS)
     NotificationPresenterClientQt::notificationPresenter()->removeClient();
 #endif
+#if ENABLE(DEVICE_ORIENTATION)
+    delete m_deviceMotionClient;
+    delete m_deviceOrientationClient;
+#endif
 }
 
 void QWebPageAdapter::deletePage()
@@ -337,6 +345,13 @@ QWebPageAdapter::VisibilityState QWebPageAdapter::visibilityState() const
 #endif
 }
 
+void QWebPageAdapter::setPluginsVisible(bool visible)
+{
+    if (!page)
+        return;
+    page->pluginVisibilityChanged(visible);
+}
+
 void QWebPageAdapter::setNetworkAccessManager(QNetworkAccessManager *manager)
 {
     if (manager == networkManager)
@@ -468,9 +483,7 @@ void QWebPageAdapter::adjustPointForClicking(QMouseEvent* ev)
     if (!foundClickableNode)
         return;
 
-    QMouseEvent* ret = new QMouseEvent(ev->type(), QPoint(adjustedPoint), ev->globalPos(), ev->button(), ev->buttons(), ev->modifiers());
-    delete ev;
-    ev = ret;
+    *ev = QMouseEvent(ev->type(), QPoint(adjustedPoint), ev->globalPos(), ev->button(), ev->buttons(), ev->modifiers());
 #else
     Q_UNUSED(ev);
 #endif
@@ -481,6 +494,8 @@ void QWebPageAdapter::mouseMoveEvent(QMouseEvent* ev)
     WebCore::Frame* frame = mainFrameAdapter()->frame;
     if (!frame->view())
         return;
+    if (ev->buttons() == Qt::NoButton)
+        mousePressed = false;
 
     bool accepted = frame->eventHandler()->mouseMoved(convertMouseEvent(ev, 0));
     ev->setAccepted(accepted);
@@ -507,7 +522,7 @@ void QWebPageAdapter::mousePressEvent(QMouseEvent* ev)
     PlatformMouseEvent mev = convertMouseEvent(ev, 1);
     // ignore the event if we can't map Qt's mouse buttons to WebCore::MouseButton
     if (mev.button() != NoButton)
-        accepted = frame->eventHandler()->handleMousePressEvent(mev);
+        mousePressed = accepted = frame->eventHandler()->handleMousePressEvent(mev);
     ev->setAccepted(accepted);
 
     RefPtr<WebCore::Node> newNode;
@@ -563,6 +578,9 @@ void QWebPageAdapter::mouseReleaseEvent(QMouseEvent *ev)
         accepted = frame->eventHandler()->handleMouseReleaseEvent(mev);
     ev->setAccepted(accepted);
 
+    if (ev->buttons() == Qt::NoButton)
+        mousePressed = false;
+
     handleSoftwareInputPanel(ev->button(), QPointF(ev->pos()).toPoint());
 }
 
@@ -1362,6 +1380,10 @@ QWebPageAdapter::ViewportAttributes QWebPageAdapter::viewportAttributesForSize(c
     return result;
 }
 
+void QWebPageAdapter::setDevicePixelRatio(float devicePixelRatio)
+{
+    page->setDeviceScaleFactor(devicePixelRatio);
+}
 
 bool QWebPageAdapter::handleKeyEvent(QKeyEvent *ev)
 {
